import java.util.*;


class Bird {
  
  class ClosestNeighboursComparator implements Comparator{
    Vector sourcePosition;
    ClosestNeighboursComparator(Vector sourcePosition){
      this.sourcePosition = sourcePosition;
    }
    int compare(Object thisBird,Object otherBird){
      float distanceToThisBird = sourcePosition.distanceTo(((Bird)thisBird).position);
      float distanceToOtherBird = sourcePosition.distanceTo(((Bird)otherBird).position);
      return new Float(distanceToThisBird).compareTo(new Float(distanceToOtherBird));
    }
  }
  
  float birdHeight = 30;
  float birdwidth = 40;
  int bodyFill = 153;
  int wingFill = 204;  
  float ang = 0;
  float flapSpeed = 0.8;
  // Properties
  Vector position;
  Vector velocity;  
  ArrayList neighbours;
  Bird[] predators;
  
  Bird(Bird[] predators){
    this.predators = predators;
    if(predators==null){
      bodyFill= color(255,5,10);
      wingFill = color(255,5,10);
    }
      
    this.ang = ((float)random(-100,100))/100.0;
    this.neighbours = new ArrayList();
    this.position = new Vector(random(-9000,9000),random(-9000,9000),random(-9000,0));
    this.velocity = new Vector(0.0,0.0,0.0);
  }
  
  void sortNeighbours(){
    Collections.sort(neighbours,new ClosestNeighboursComparator(position));
  }
   
   //separation
  Vector distanceVelocity(){
    float minimumDistance = 400;
    Vector distanceVelocity = new Vector(0,0,0);
    for(int i=0;i<30;i++){
      Bird neighbour = (Bird) neighbours.get(i);
      if(position.distanceTo(neighbour.position) < minimumDistance)
          distanceVelocity = distanceVelocity.subtract(neighbour.position.subtract(this.position));
    }
    distanceVelocity = distanceVelocity.devide(30);
    return distanceVelocity.normalize();
  }
  
  Vector attackClosestVelocity(){
    float minimumDistance = 800;
    Vector distanceVelocity = new Vector(0,0,0);
    Bird neighbour = (Bird) neighbours.get(0);
    if(position.distanceTo(neighbour.position) < minimumDistance)
          distanceVelocity = distanceVelocity.add(neighbour.position.add(this.position));
    
    return distanceVelocity.normalize();
  }
  
  
  //cohesion
  Vector centerVelocity(){
    Vector centerVelocity = new Vector(0,0,0);
    for(int i=0;i<neighbours.size();i++){      
      Bird neighbour = (Bird) neighbours.get(i % 5 == 0 ? (int) random(0,10) : i);
      centerVelocity = centerVelocity.add(neighbour.position);
    }
    centerVelocity = centerVelocity.devide(100);
    return centerVelocity.subtract(position).normalize();
  }
  
  Vector centerOfBoxVelocity(){
    Vector centerVelocity = new Vector(0,0,0);
    return centerVelocity.subtract(position).normalize();
  }
  
  Vector avoidPredators(){
    Vector avoidPredatorVelocity = new Vector(0.0,0.0,0.0);
    for(int p=0;p < predators.length ; p++){
      Bird predator = predators[p];
      Vector predatorPlace = predator.position;
      if(predator.position.distanceTo(position) < 3000)
        return position.subtract(predator.position).normalize();
    }
    return avoidPredatorVelocity;
  }
  
  //alignment
  Vector similarVelocity(){
    Vector similarVelocity = new Vector(0.0,0.0,0.0);
    for(int i=0;i<5;i++){
      Bird neighbour = (Bird) neighbours.get(i);
      similarVelocity = similarVelocity.add(neighbour.velocity);
    }
    similarVelocity = similarVelocity.devide(5);
    return similarVelocity.subtract(velocity).normalize();
  }
  
  Vector limitVelocity(Vector velocity){
    float maxSpeed = predators == null ? 440 : 400;
    float speed = velocity.size();
    if(speed > maxSpeed)
      return velocity.devide(speed).multiply(maxSpeed);
    return velocity;
  }
  
  Vector boundingBoxVelocity(){
    int maxX = width * 30;
    int maxY = height * 30;
    int maxZ = height * 30;
    int force = 70;
    
    float velocity_x = 0.0;    
    if(position.x < -1 * maxX)
      velocity_x = force;
    else if(position.x > maxX)
      velocity_x =  force * -1;
      
    float velocity_y = 0.0;
    if(position.y < -1 * maxY)
      velocity_y = force ;
    else if(position.y > maxY)
      velocity_y = force * -1;      
      
    float velocity_z = 0.0;
    if(position.z < -1 * maxZ)
      velocity_z = force;
    else if(position.z > maxZ)
      velocity_z =  -1 * force;
    
    return new Vector(velocity_x,velocity_y,velocity_z).normalize();
  }
    
  void addNeighbour(Bird otherBird){
    this.neighbours.add(otherBird);
  }


  void fly() {
    pushMatrix();
    
    try{
    //Thread.sleep(20);
    }catch(Exception e){
    }
    
    sortNeighbours();
    /*
     System.out.println("position:\t\t" + position);  
    System.out.println("bounding:\t\t" +boundingBoxVelocity());
    System.out.println("center v:\t\t" +centerVelocity().devide(1000));
    System.out.println("distancev:\t\t" + distanceVelocity());
    */
    if(predators == null){
      velocity = limitVelocity(
      attackClosestVelocity().multiply(55)
      .add(centerVelocity().multiply(40))
      );
    }else{
      velocity = limitVelocity(
     boundingBoxVelocity().multiply(56)
    .add(centerVelocity().multiply(25))
    .add(distanceVelocity().multiply(55))
    .add(similarVelocity().multiply(45))
    .add(avoidPredators().multiply(48))
    .add(centerOfBoxVelocity().multiply(25))
    );
/*
    System.out.println("predatorp:\t\t" + predator.position);    
    System.out.println("predatorv:\t\t" + avoidPredator());
    System.out.println("similar v:\t\t" + similarVelocity());    
    System.out.println("total v:\t\t" + velocity);
 */ 
    
    }
    
    position = position.add(velocity);
    //System.out.println("new position:\t" + position + "\n");
    
    translate(position.x,position.y,position.z-5000);
    
    

    rotateX(PI/8);
    rotateY(PI/2);//
    rotateZ(PI/2);//position.getZAngle()

    if(predators == null){
     // System.out.println("new position:\t" + position + "\n");
    // Body
    fill(bodyFill);
    box(20,100,20);
    
    //System.out.println(PI/4);
    //System.out.println(new Vector(1.0,1.0,0.0).getXAngle());
    //System.out.println(new Vector(1.0,1.0,0.0).getYAngle());
    //System.out.println(new Vector(1.0,1.0,0.0).getZAngle());

    // Left wing
    fill(wingFill);
    pushMatrix();
    rotateY(sin(radians(ang)) * 20);
    rect(-75, -50, 75, 100);
    popMatrix();

    // Right wing
    pushMatrix();
    rotateY(sin(radians(ang)) * -20);
    rect(-75, -50, 75, 100);
    popMatrix();
    
    }else{
      
    // Body
    fill(bodyFill);
    box(40,200,40);

    // Left wing
    fill(wingFill);
    pushMatrix();
    rotateY(sin(radians(ang)) * 20);
    rect(-150, -100, 150, 200);
    popMatrix();

    // Right wing
    pushMatrix();
    rotateY(sin(radians(ang)) * -20);
    rect(-150, -100, 150, 200);
    popMatrix();  
    }

    // Wing flap
    double factor = 1;
    if (ang > 2) {
      factor*=-1;
    } 
    if (ang < -2) {
      factor*=-1;
    }

    ang += (float)velocity.size() / 50.0  ;
    
    popMatrix();
  }
}




